(in-package :swank)

(defun levenshtein-distance (source target &key (test #'char=))
  (block nil
    (let ((source-length (length source))
	  (target-length (length target)))
      (when (zerop source-length)
	(return target-length))
      (when (zerop target-length)
	(return source-length))
      (let ((buffer (make-array (1+ target-length))))
	(dotimes (i (1+ target-length))
	  (setf (aref buffer i) i))
	;; we make a slight modification to the alogrithm. we don't
	;; create the entire matrix, just enough to keep the info we
	;; need, which is an array of size target-length + the "above"
	;; value and the "over". (this is similar to the optimizaiont
	;; for determining lcs).
	(loop
	   for i from 1 upto source-length
	   do (setf (aref buffer 0) i)
	   do (loop
		 with above-value = i
		 with over-value = (1- i)
		 for j from 1 upto target-length
		 for cost = (if (funcall test (aref source (1- i))
					      (aref target (1- j)))
				0 1)
		 do (let ((over-value* (aref buffer j)))
		      (setf (aref buffer j) (min (1+ above-value)
						 (1+ (aref buffer j))
						 (+ cost over-value))
			    above-value (aref buffer j)
			    over-value over-value*))))
	(return (aref buffer target-length))))))

(defmacro xor (a b)
  `(if ,a
       (not ,b)
       ,b))

(defun ensure-list (obj)
  (if (listp obj)
      obj
    (list obj)))

(defun extended-apropos-list (string &key (fbound :dont-care)
                                          (bound :dont-care)
                                          (package :dont-care)
                                          (external t)
                                          (distance nil))
  "Returns a list of symbols. If FBOUND is T only fbound symbols
  will be returned, if it is NIL only not fbound symbols will be
  returned, if it is :DONT-CARE both fbound and not fbound
  symbols will be returned. Idem for bound, package and external.

  DISTANCE can be nil in which case all symbols whose name
  contains STRING will be returned, or a number specifying the
  edit distance of the symbols we want. NB: specifying a distance
  argument elminates the sub-string check."
  (let ((symbols '())
        (package-list (ensure-list package)))
    (do-all-symbols (sym)
      (block collect-symbol

        (unless (eql :dont-care external)
          (when (xor external (symbol-external-p sym))
            (return-from collect-symbol)))

        (unless (eql :dont-care fbound)
          (when (xor fbound (fboundp sym))
            (return-from collect-symbol)))

        (unless (eql :dont-care bound)
          (when (xor bound (boundp sym))
            (return-from collect-symbol)))

       (unless (eql :dont-care package)
         (unless (member (symbol-package sym) package-list :test #'eq)
           (return-from collect-symbol)))

       (etypecase distance
         (number
          (unless (and
                   (<= (abs (- (length (symbol-name sym)) 
                               (length string)))
                       distance)
                   (<= (levenshtein-distance string (symbol-name sym)) distance))
            (return-from collect-symbol)))
         ((or (eql nil) (eql :dont-care))
          (unless (search string (symbol-name sym) :test #'char-equal)
            (return-from collect-symbol))))

       ;; all the checks we wanted to perform passed.
       (push sym symbols)))
    symbols))

(defslimefun extended-apropos-list-for-emacs (&rest extended-apropos-args)
  (mapcan (listify #'briefly-describe-symbol-for-emacs)
          (sort (apply #'extended-apropos-list extended-apropos-args)
                #'present-symbol-before-p)))
